# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import collections
import copy
import dataclasses as dc
import importlib
import types
import typing as tp

from .._logging import logger
from .._proto_namespace import _ProtoNamespace
from .._yaml import yaml  # type: ignore
from ._05_project import ProjectMixin
from ._payload import DockerBuildPayload
from ._payload import DockerComposePayload
from ._payload import DockerRunPayload
from ._payload import Payload
from ._payload import PythonPayload
from ._payload import WithDependencyPayload


class PayloadMixin(ProjectMixin):
    @classmethod
    def __keys(cls) -> tp.Tuple[str, ...]:
        return ("docker_settings",) + tuple(cls.payload_types())

    @classmethod
    def payload_types(cls) -> tp.Dict[str, tp.Type[Payload]]:
        type_list: tp.List[tp.Type[Payload]] = [
            DockerBuildPayload,
            DockerRunPayload,
            PythonPayload,
            DockerComposePayload,
            WithDependencyPayload,  # must be at end for now
        ]
        return {x.flavor: x for x in type_list}

    @classmethod
    def level_map(cls) -> tp.Dict[str, int]:
        res = {}
        for flavor, val in cls.payload_types().items():
            res[flavor] = 0
            if val == WithDependencyPayload:
                res[flavor] = 1
        return res

    def templated(self) -> None:
        super().templated()
        self.auxcon.payload = _ProtoNamespace()

    def formatted(self) -> None:
        super().formatted()
        self._copy_keys_over(self.__keys(), "payload")
        for key in self.__keys():
            if key in ["docker_settings"]:
                continue  # handled by docker
            self._to_proto_ns("payload", key, iter_mapping=True)

        try:
            for val in self.auxf.payload.with_dependency.values():
                try:
                    if isinstance(val.deps, str):
                        val.deps = [val.deps]
                except KeyError:
                    pass
        except KeyError:
            pass

    def defaulted(self) -> None:
        super().defaulted()
        self.auxd.setdefault("payload", _ProtoNamespace())
        data = self.auxd.payload
        for key in self.__keys():
            if key in ["docker_settings"]:
                continue  # handled by docker
            data.setdefault(key, _ProtoNamespace())

        prefix = "build-"
        for payload_name, val in data.with_dependency.items():
            # param gets removed in enriched
            if payload_name.startswith(prefix):
                val.setdefault("deps", [])
            else:
                val.setdefault("deps", [f"{prefix}{payload_name}"])

            val.deps = val.deps or []
            val.setdefault("payload", payload_name)

    def enriched(self) -> None:
        super().enriched()
        data = self.auxe.payload

        self._add_absend_and_expand_variants(data)

        # helper parameter
        for flavor in self.payload_types():
            for payload_name, payload in data[flavor].items():
                payload.name = payload_name
                payload.flavor = flavor

        lookup_cfg: DoubleDict[_ProtoNamespace] = DoubleDict(self.level_map())
        for flavor in list(self.payload_types()):
            lookup_cfg.update(data[flavor], unique=True, flavor=flavor)

        for _, val in lookup_cfg.level("with_dependency").items():
            try:
                # we can specify the payload with flavor:name,
                # but if it is only name, the flavor will get deduced
                if ":" in val.payload:
                    pflv, pname = val.payload.split(":", 1)
                    # produce key error if not available
                    data[pflv][pname]  # pylint: disable=pointless-statement
                else:
                    payload_cfg = lookup_cfg[(val.payload, val.name, 1)]
                    val.payload = f"{payload_cfg.flavor}:{payload_cfg.name}"
            except KeyError as err:
                raise RuntimeError(
                    f"did not find matching payload for '{val.name}'. Ensure it exists or use the payload argument."
                ) from err

    def _add_absend_and_expand_variants(self, data: _ProtoNamespace) -> None:
        # resolve variants
        for flavor in self.payload_types():
            # we should only _add_if_absent after the lower levels
            # had a change to expand their variants
            if flavor == "with_dependency":
                for payload_name, val in data.with_dependency.items():
                    param = val.get("param", _ProtoNamespace())
                    for dep in val.deps:
                        self._add_if_absent(dep, param, data)

                    val.setdefault("payload", payload_name)
                    self._add_if_absent(val.payload, param, data)

            for service_name, val in list(data[flavor].items()):
                self._expand_variant(flavor, service_name, val)

        for payload_name, val in data.with_dependency.items():
            val.pop("param", _ProtoNamespace())

    def _add_if_absent(
        self,
        payload_name: str,
        param: _ProtoNamespace,
        data: _ProtoNamespace,
        after: tp.Optional[str] = None,
    ) -> None:
        # prevents having to write a service in docker_run,
        # docker_build, and in with_dependency
        check_flavors = [flv for flv, val in self.level_map().items() if val == 0]
        flavor = DockerRunPayload.flavor
        if payload_name.startswith("build-"):
            flavor = DockerBuildPayload.flavor

        if all(payload_name not in data[flv] for flv in check_flavors):
            val = _ProtoNamespace(param=copy.deepcopy(param))
            if after:
                data[flavor].insert_after(after, payload_name, val)
            else:
                data[flavor][payload_name] = val
            logger.info("added %s:%s automatically", flavor, payload_name)

    def _expand_variant(
        self,
        flavor: str,
        base_service_name: str,
        base_val: _ProtoNamespace,
        interfix: str = "",
        prefix_opts: tp.Optional[_ProtoNamespace] = None,
    ) -> None:
        data = self.auxe.payload
        base_variants = base_val.pop("variant", _ProtoNamespace())

        # sanatize base_variants
        base_variants = _ProtoNamespace(
            (interfix + key, val or _ProtoNamespace())
            for key, val in base_variants.items()
        )

        if not prefix_opts:
            # prefix opts allows us to reuse the settings of a
            # variant later again, i.e.
            # variant:
            #     -mr:
            #         bla: 1
            #     -39:
            #         version: "3.9"
            #         variant:
            #             -mr:
            #             # has automatically bla: 1
            prefix_opts = _ProtoNamespace(
                {interfix: base_val.get("param", _ProtoNamespace())}
            )

        # prepare the param object for the variants
        for fvariant, mod in base_variants.items():
            if interfix:
                variant = fvariant.split(interfix, 1)[1]
            else:
                variant = fvariant
            prefix_opts[fvariant] = copy.deepcopy(prefix_opts[interfix])
            if variant in prefix_opts and interfix != "":
                prefix_opts[fvariant].update(prefix_opts[variant])
                logger.info(
                    "%s -> applied %s %s variant options to %s variant",
                    base_service_name,
                    variant,
                    prefix_opts[variant],
                    fvariant,
                )
            prefix_opts[fvariant].update(mod)

        # create the variants
        for fvariant, mod in reversed(base_variants.items()):
            # do subvariants first
            subvariants = mod.pop("variant", _ProtoNamespace())
            base_val_mod = copy.deepcopy(base_val)
            base_val_mod.param = copy.deepcopy(prefix_opts[fvariant])
            base_val_mod.variant = subvariants
            self._expand_variant(
                flavor,
                base_service_name,
                base_val_mod,
                interfix=fvariant,
                prefix_opts=prefix_opts,
            )

            # then self and potential automatic payload
            val = copy.deepcopy(base_val)
            val.setdefault("param", _ProtoNamespace())
            val.param = prefix_opts[fvariant]
            val.param.pop("variant", None)

            if flavor == "with_dependency":
                param = val.param
                param.setdefault("docker_name", base_val.payload)
                param.image_name = self.auxe.project.slug + "-" + base_val.payload
                val.payload += fvariant
                self._add_if_absent(val.payload, param, data, base_service_name)
            if flavor == "python":
                val.setdefault("fct", base_service_name)

            # finally add it
            data[flavor].insert_after(
                base_service_name, base_service_name + fvariant, val
            )

    def hydrated(self) -> None:
        super().hydrated()
        # hydration
        lookup: DoubleDict[Payload] = DoubleDict(self.level_map())
        type_sel = self.payload_types()
        data = self.auxh.payload
        custom_ppyload = self._load_custom_ppyload()

        def create_payload(kwgs: _ProtoNamespace, flavor: str) -> Payload:
            kwgs.auxh = self.auxh
            if flavor == "python":
                kwgs.fct = PythonPayload.import_function(
                    kwgs.get("fct", kwgs.name), custom_ppyload
                )
            if flavor == "with_dependency":
                kwgs.deps = tuple(lookup[key] for key in kwgs.deps)
                assert all(isinstance(x, Payload) for x in kwgs.deps)
                kwgs.payload = lookup[kwgs.payload]
            try:  # see if it exists...
                res = lookup[kwgs.name]
                if res.flavor != flavor:
                    raise KeyError(kwgs.name)
            except KeyError:  # ...if not, create
                # remove helper variable
                kwgs.pop("flavor")
                res = type_sel[flavor](**kwgs)

            assert isinstance(res, Payload)
            lookup.update({key: res}, flavor=flavor, unique=True)
            return res

        for flavor in self.payload_types():
            for key, val in data[flavor].items():
                create_payload(val, flavor)
        data.lookup = lookup
        for val in lookup.level("with_dependency").values():
            val.hydrate()
        for val in lookup.level("docker_build").values():
            val.hydrate()

    def payload_run(self, *payloads: Payload, force: bool, dry: bool = False) -> bool:
        # pylint: disable=import-outside-toplevel
        from ._aux_ci import JobManager, CommonJob

        @dc.dataclass(frozen=True)
        class _PayloadJobImpl:
            payload: Payload
            force: bool

        @dc.dataclass(frozen=True)
        class PayloadJob(CommonJob, _PayloadJobImpl):
            def is_up_to_date(self, env: tp.Mapping[str, tp.Any]) -> bool:
                if self.force:
                    return False
                return self.payload.is_up_to_date()

            def script(self) -> None:
                self.payload.run(self.force)

            def is_included(self, env: tp.Mapping[str, tp.Any]) -> tp.Tuple[bool, str]:
                if isinstance(self.payload, DockerBuildPayload):
                    if self.payload.is_up_to_date() and not self.force:
                        return False, f"{self.payload.executor.tag} exists"
                    else:
                        return True, f"missing {self.payload.executor.tag}"

                return super().is_included(env)

        jman = JobManager(env=dict(verbose=True))

        def demo(*pld: Payload) -> tp.List[PayloadJob]:
            jobs = []
            for pl in pld:
                if isinstance(pl, WithDependencyPayload):
                    parents = demo(*pl.deps)
                    name = pl.payload.name
                    if name in jman.jobs:
                        x = jman.jobs[name]
                    else:
                        x = jman.add_job(
                            PayloadJob,
                            payload=pl.payload,
                            force=force,
                            name=pl.payload.name,
                            parents=parents,
                        )
                    jobs.append(x)
                else:
                    name = pl.name
                    if name in jman.jobs:
                        x = jman.jobs[name]
                    else:
                        x = jman.add_job(
                            PayloadJob, payload=pl, force=force, name=pl.name
                        )
                    jobs.append(x)
            return jobs  # type: ignore

        demo(*payloads)

        if dry:
            jman.prune_unincluded(show=True)
        else:
            jman.run_pipeline()
            jman.show_result()

        if jman.has_failed_jobs():
            return False
        return True

    def _load_custom_ppyload(self) -> tp.Optional[types.ModuleType]:
        try:
            with self.preprend_to_sys_path(self.target / "custom/"):
                custom_ppyload = importlib.import_module("payload.python.functions")
                logger.info("custom python payloads found")
        except (ImportError, FileNotFoundError):
            custom_ppyload = None
            logger.info("custom python payloads NOT found")
        return custom_ppyload


X = tp.TypeVar("X")


@dc.dataclass
class DoubleDict(tp.Generic[X]):
    level_map: tp.Dict[str, int]
    level_dict: tp.Dict[int, tp.Dict[str, X]] = dc.field(
        default_factory=lambda: collections.defaultdict(dict)
    )

    def update(self, rhs: tp.Dict[str, X], flavor: str, unique: bool = False) -> None:
        level = self.level_map[flavor]
        if unique:
            lhs_keys = set(self.level_dict[level])
            rhs_keys = set(rhs)
            overlap = lhs_keys & rhs_keys
            if overlap:
                raise RuntimeError(f"keys {overlap} are not unique!")

        self.level_dict[level].update(rhs)

    def __getitem__(self, key_or_tuple: tp.Union[str, tp.Tuple[str, str, int]]) -> X:
        skip_level = -1  # -1: means nothing is skipped, first hit is returned
        if isinstance(key_or_tuple, tuple):
            key, reduce_match, skip_level = key_or_tuple
            assert ":" not in key
            assert ":" not in reduce_match
            if key != reduce_match:
                skip_level = -1
        else:
            key = key_or_tuple

        if ":" in key:
            flavor, key = key.split(":")
            level = self.level_map[flavor]
            return self.level_dict[level][key]

        for level in reversed(sorted(self.level_dict)):
            if key in self.level_dict[level]:
                if level != skip_level:
                    return self.level_dict[level][key]

        raise KeyError(key)

    def level(self, flavor: str) -> tp.Dict[str, X]:
        level = self.level_map[flavor]
        return self.level_dict[level]


def repr_double_dict(
    representer: tp.Any, data: tp.Any  # pylint: disable=unused-argument
) -> tp.Any:
    return representer.represent_str("n/a")


yaml.representer.add_representer(DoubleDict, repr_double_dict)



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
# pylint: disable=too-many-lines
from ._15_pytest import PytestMixin


class CoverageMixin(PytestMixin):
    def templated(self) -> None:
        super().templated()
        self.auxcon.dependencies.test.append(self.versions.pytest_cov)



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
# pylint: disable=too-many-lines
import typing as tp

from .._parser import Jinja2Parser
from .._parser import YamlParser
from .._proto_namespace import _ProtoNamespace
from ._03_meta import MetaMixin
from ._05_project import ProjectMixin
from ._06_dependency import DependencyMixin
from ._18_payload import PayloadMixin


class DockerMixin(PayloadMixin, DependencyMixin, ProjectMixin, MetaMixin):
    @classmethod
    def __keys(cls) -> tp.Tuple[str, ...]:
        return ("base_match", "platform", "compose_version")

    def templated(self) -> None:
        super().templated()
        data = self.auxcon.payload

        if self.is_enabled("pre-commit"):
            data.setdefault("with_dependency", _ProtoNamespace())
            run = "pre-commit"
            build = f"build-{run}"
            data.with_dependency[run] = dict(
                deps=[build], variant={"-all": dict(cmd="--all")}
            )
            data.with_dependency[build] = dict(deps=["build-python-deps"])

        if self.is_enabled("pytest"):
            data.setdefault("with_dependency", _ProtoNamespace())
            run = "pytest"
            build = f"build-{run}"
            data.with_dependency[build] = dict(deps=["build-python-deps"])
            data.with_dependency[run] = dict(deps=[build])
            if self.is_enabled("coverage"):
                data.with_dependency[run]["variant"] = {
                    "-mr": {"marker": ""},
                    "-cov": dict(coverage=95, variant={"-mr": None}),
                }

        if self.is_enabled("docs"):
            data.setdefault("with_dependency", _ProtoNamespace())
            data.with_dependency["docs"] = dict(param=dict(extra_req="docs"))

        if self.is_enabled("gitlab"):
            data.setdefault("docker_run", _ProtoNamespace())
            data.setdefault("python", _ProtoNamespace())
            data.docker_run["gitlab-release-run"] = _ProtoNamespace()
            data.python["gitlab-release"] = _ProtoNamespace()
            data.docker_run["pkg-gitlab"] = _ProtoNamespace()

    def demodata(self) -> None:
        super().demodata()
        data = self.auxcon.payload
        data.docker_settings = _ProtoNamespace(platform="amd64")

    def update_to_template(self, tpl: _ProtoNamespace, full: _ProtoNamespace) -> None:
        super().update_to_template(tpl, full)

        for flavor in ["docker_build", "docker_run", "with_dependency"]:
            if flavor not in self.auxf.payload:
                continue
            if flavor not in tpl.payload:
                continue
            data = self.auxf.payload[flavor]
            old = list(data)
            new = list(tpl.payload[flavor])
            last_idx = 0
            for name, payload in full.payload[flavor].items():
                if name in new and name not in old:
                    data[name] = payload
                    old.insert(last_idx + 1, name)
                    self._print(f"payload.{flavor}: added {name}", fg="green")
                elif name in old and name not in new:
                    del data[name]
                    old.remove(name)
                    self._print(f"payload.{flavor}: removed {name}", fg="red")
                if name in old and name in new:
                    last_idx = old.index(name)

    def formatted(self) -> None:
        super().formatted()
        self._copy_keys_over(self.__keys(), "payload", "docker_settings")
        self._to_proto_ns("payload", "docker_settings")
        self._to_proto_ns("payload", "docker_settings", "base_match", iter_mapping=True)

    def defaulted(self) -> None:
        super().defaulted()
        payload_data = self.auxd.payload
        payload_data.setdefault("docker_settings", _ProtoNamespace())
        data = payload_data.docker_settings
        data.setdefault("base_match", _ProtoNamespace())
        data.setdefault("platform", None)
        data.setdefault("compose_version", self.versions.docker_compose_file)

    def enriched(self) -> None:
        super().enriched()
        data = self.auxe.payload.docker_settings
        data.project_dir = "../.."
        data.source_dir = f"../../{self.auxe.project.source_dir}"
        self._docker_build_enrich()
        self._docker_run_enrich()

    def _docker_build_enrich(self) -> None:
        extra_req_default = {
            "python-deps": ["default"],
            "pytest": ["test"],
            "pre-commit": ["test", "dev"],
            "pytest-standalone": ["default", "test"],
            "docs": ["docs"],
            "ansible-deploy": ["deploy"],
        }
        deps = self.auxe.dependencies
        data = self.auxe.payload
        prefix = "build-"
        for key, payload_cfg in data.docker_build.items():
            assert key.startswith(prefix)
            name = key[len(prefix) :]
            payload_cfg.setdefault("param", _ProtoNamespace())
            param = payload_cfg.param
            if "extra_req" in param and isinstance(param.extra_req, str):
                param.extra_req = [param.extra_req]

            param.setdefault("pip_req", [])
            param.setdefault("script", [])
            param.setdefault("base", None)
            self._adjust_base_on_match_entry(param, self.auxe)

            param.setdefault("apt_req", [])

            self._set_name_and_version(param, name)

            if "mode" in param:
                supported = ["django", "django+nginx"]
                if param.mode not in supported:
                    raise RuntimeError(
                        f"mode {param.mode} is not supported {supported}"
                    )

            # branch matching and dep settings
            fallback = extra_req_default.get(param.docker_name, [])
            needed = param.get("extra_req", fallback)
            param.pip_req = self._unique_sum(
                param.pip_req, *[deps.get(x, []) for x in needed]
            )
            param.apt_req = self._unique_sum(
                param.apt_req, *[deps.get(x + "_apt", []) for x in needed]
            )
            param.script = self._sum(
                param.script, *[deps.get(x + "_script", []) for x in needed]
            )
            assert all('"' not in x for x in param.pip_req)

            if self.is_enabled("pip"):
                self.branch_match_and_cred_passing(param, self.auxe)  # type: ignore

    def _docker_run_enrich(self) -> None:
        data = self.auxe.payload
        for key, payload_cfg in data.docker_run.items():
            payload_cfg.setdefault("param", _ProtoNamespace())
            param = payload_cfg.param

            res = []
            if "assets" in param:
                if isinstance(param.assets, str):
                    param.assets = [param.assets]

                for x in param.assets:
                    shortname = x.rsplit("/", 1)[1]
                    url = "$CI_API_V4_URL/projects/$CI_PROJECT_ID/packages/generic/" + x
                    res.append(
                        r"{\"name\":\"" + shortname + r"\",\"url\":\"" + url + r"\"}"
                    )
            param.assets = res
            self._set_name_and_version(param, key)
            param.setdefault("services", [param.service_name])

    def _set_name_and_version(self, param: _ProtoNamespace, name: str) -> None:
        slug = self.auxe.project.slug
        param.service_name = name
        if param.service_name == slug:
            raise RuntimeError(
                f"service name should not be identical to project slug '{slug}'"
            )

        param.setdefault("image_name", f"{slug}-{param.service_name}")

        param.setdefault("docker_name", name)
        param.setdefault("version", self.auxe.project.minimal_version)
        if not isinstance(param.version, str):
            raise RuntimeError(
                f"please specity the version '{param.version}' in {name} as a string"
            )
        param.setdefault("version_slug", param.version.replace(".", ""))

    def _adjust_base_on_match_entry(
        self,
        opts: _ProtoNamespace,
        auxe: _ProtoNamespace,
    ) -> None:
        base_match = auxe.payload.docker_settings.base_match
        if opts.base is None:
            return
        image, tag = opts.base.rsplit(":", 1)
        if image not in base_match:
            return
        fallback = base_match[image].fallback
        var = (
            f'{image.upper().replace("/", "_").replace(".", "_").replace("-", "_")}_TAG'
        )
        skip = [auxe.gitlab.release_branch]

        # changing this order affects the jinja2 files!
        opts.base_match = [(image, var, skip, fallback, tag)]

        opts.base = f"{image}:${var}"

    def bake(self) -> None:  # pylint: disable=too-many-branches
        super().bake()
        data = self.auxe.payload
        settings = self.auxe.payload.docker_settings

        config = _ProtoNamespace(
            [("version", settings.compose_version), ("services", {})]
        )

        # pylint: disable=too-many-nested-blocks
        for val in data.docker_run.values():
            self._bake_docker_compose_file(val, config)

        for val in data.docker_build.values():
            self._bake_docker_build_file(val)
            self._bake_docker_compose_file(val, config)

        dest = self.target / "docker/compose.yml"
        written = YamlParser.write(config, dest)
        if written:
            self._print(f"baked {dest}", fg="green")

        # self._add_compositions(config)

    def _bake_docker_build_file(self, payload: _ProtoNamespace) -> None:
        """
        Return true if the file was a custom one.
        """
        opts = payload.param
        try:
            self.bake_file(
                f"docker/services/{opts.docker_name}/Dockerfile",
                f"docker/{opts.service_name}.dockerfile",
                opts=opts,
            )
            return
        except FileNotFoundError:  # then it must be custom
            pass

        self.bake_file(
            f"docker/{opts.docker_name}.dockerfile",
            f"docker/{opts.service_name}.dockerfile",
            custom=True,
            opts=opts,
        )

    def _bake_docker_compose_file(
        self, payload: _ProtoNamespace, config: _ProtoNamespace
    ) -> None:
        opts = payload.param
        src_dir = self.root / f"docker/services/{opts.docker_name}"
        srcj = src_dir / "docker-compose.yml.jinja2"
        custom_srcj = self.target_custom / "docker/compose.yml.jinja2"
        custom_src = self.target_custom / "docker/compose.yml"

        if srcj.exists():
            with Jinja2Parser.render_to_tmp(srcj, aux=self.auxe, opts=opts) as src:
                part = YamlParser.read(src).services
                self._update_config(payload, part, config)
                return
        elif custom_srcj.exists():
            with Jinja2Parser.render_to_tmp(
                custom_srcj, aux=self.auxe, opts=opts
            ) as src:
                custom_config = YamlParser.read(src)
        else:
            if not custom_src.exists():
                raise RuntimeError(
                    f"{custom_src} not found! Needed by {opts.docker_name}."
                )
            custom_config = YamlParser.read(custom_src)

        if list(custom_config.keys()) != ["services"]:
            raise RuntimeError(
                "only services can be defined in custom docker-compose.yml."
            )
        part = custom_config.services
        self._update_config(payload, part, config)

    def _update_config(
        self, payload: _ProtoNamespace, part: _ProtoNamespace, config: _ProtoNamespace
    ) -> None:
        for service_name, val in part.items():
            for key in list(val):
                if payload.flavor == "docker_build":
                    if key not in ["build", "image"]:
                        del val[key]
                elif payload.flavor == "docker_run":
                    if key in ["build"]:
                        del val[key]
                else:
                    raise NotImplementedError(payload.flavor)
            config["services"].setdefault(service_name, val)
            config["services"][service_name].update(val)

    # def _add_compositions(self, compose_all_config: _ProtoNamespace) -> None:
    #     data = self.auxe.docker
    #     for comp_name, comp in data.get("compositions", {}).items():
    #         # check validity and set defaults
    #         valid_keys = {"pre_build", "pre_script", "services", "overwrite"}
    #         invalid_keys = set(comp.keys()) - valid_keys
    #         if invalid_keys:
    #             raise RuntimeError(
    #                 f"invalid keys {invalid_keys} in composition '{comp_name}'. "
    #                 f"Valid key are {valid_keys}."
    #             )
    #         for key in valid_keys:
    #             comp.setdefault(key, [])
    #         comp.name = comp_name
    #         prefix = self.auxe.project.slug
    #         comp.services = [f"{prefix}-{key}" for key in comp.services]
    #         comp.pre_build = [f"{prefix}-{key}" for key in comp.pre_build]
    #         og_overwrite = comp.overwrite
    #         comp.overwrite = [f"../custom/docker/{key}" for key in comp.overwrite]
    #         for overwrite, file in zip(og_overwrite, comp.overwrite):
    #             search = (self.target / "docker" / file).resolve()
    #             if not search.exists():
    #                 raise RuntimeError(
    #                     f"overwrite file '{overwrite}' does not exist at '{search.relative_to(self.target.resolve().parent)}'"
    #                 )

    #         # docker-compose file
    #         config = copy.deepcopy(compose_all_config)
    #         config.services = {
    #             key: val
    #             for key, val in compose_all_config.services.items()
    #             if key in comp.services
    #         }
    #         config.setdefault("networks", {})
    #         config["networks"][f"{comp.name}-network"] = {}

    #         for val in config.services.values():
    #             val.setdefault("networks", [])
    #             val["networks"].append(f"{comp.name}-network")
    #         comp.compose_filename = f"{comp.name}-compose.yml"
    #         dest = self.target / f"docker/{comp.compose_filename}"
    #         written = YamlParser.write(config, dest)
    #         if written:
    #             self._print(f"baked {dest}", fg="green")

    #         # helper bash file
    #         self.bake_file(
    #             "docker/comp-helper.sh", f"docker/{comp.name}-compose.sh", comp=comp
    #         )

    @staticmethod
    def _unique_sum(*args: tp.List[str]) -> tp.List[str]:
        res = []
        for part in args:
            for x in part:
                if x not in res:
                    res.append(x)
        return res

    @staticmethod
    def _sum(*args: tp.List[str]) -> tp.List[str]:
        res = []
        for part in args:
            for x in part:
                res.append(x)
        return res



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import configparser
import contextlib
import filecmp
import shutil
import tempfile
import typing as tp
from pathlib import Path

import jinja2

from ._base_parser import BaseParser
from ._proto_namespace import _ProtoNamespace
from ._yaml import CommentedMap  # type: ignore
from ._yaml import CommentedSeq  # type: ignore
from ._yaml import yaml  # type: ignore


class ConfigParser(BaseParser):
    @classmethod
    def to_conf(cls, data: _ProtoNamespace) -> tp.Any:
        conf = configparser.ConfigParser()
        conf.read_dict(data)
        return conf

    @classmethod
    def from_conf(cls, conf: tp.Any) -> _ProtoNamespace:
        res = _ProtoNamespace()
        for key, val in conf.items():
            if key != "DEFAULT":
                res[key] = val

        cls.rec_walk(
            res,
            [
                (
                    lambda x: isinstance(x, configparser.SectionProxy),
                    lambda x: _ProtoNamespace(x.items()),
                ),
                (
                    lambda x: isinstance(x, str) and "\n" in x,
                    lambda x: x.strip().split("\n"),
                ),
            ],
        )
        return res

    @classmethod
    def read(cls, filename: Path) -> _ProtoNamespace:
        conf = configparser.ConfigParser()
        conf.read(filename)
        return cls.from_conf(conf)

    @classmethod
    def read_string(cls, text: str) -> _ProtoNamespace:
        conf = configparser.ConfigParser()
        conf.read_string(text)
        return cls.from_conf(conf)

    @classmethod
    def write(cls, data: _ProtoNamespace, dest: Path) -> bool:
        with tempfile.NamedTemporaryFile("w") as tmp:
            with open(tmp.name, "w", encoding="utf-8") as f:
                cls.write_stream(data, f)
            cls._remove_trailing_space(Path(f.name))
            if dest.exists() and filecmp.cmp(f.name, dest):
                return False
            cls.ensure_parent(dest)
            shutil.copyfile(f.name, dest)
            return True

    @classmethod
    def write_stream(cls, data: _ProtoNamespace, f: tp.TextIO) -> None:
        conf = cls.to_conf(data)
        cls.rec_walk(
            conf,
            [
                (
                    lambda obj: isinstance(obj, str)
                    and obj.startswith("[")
                    and obj.endswith("]"),
                    lambda x: "\n"
                    + "\n".join(
                        y[1:-1] if y[0] == "'" else y for y in x[1:-1].split(", ")
                    )
                    if x[1:-1]
                    else "",
                ),
                (
                    lambda obj: isinstance(obj, str)
                    and obj.startswith("{")
                    and obj.endswith("}"),
                    lambda x: "\n"
                    + "\n".join(
                        (y[1:-1] if y[0] == "'" else y).replace("': '", " = ")
                        for y in x[1:-1].split(", ")
                    )
                    if x[1:-1]
                    else "",
                ),
            ],
        )
        conf.write(f)

    @classmethod
    def _remove_trailing_space(cls, dest: Path) -> None:
        with dest.open("r") as f:
            lines = f.readlines()

        for i, x in enumerate(lines):
            if len(x) > 1 and x[-2] == " ":
                lines[i] = x[0:-2] + "\n"

        if lines[-1] == "\n":
            lines.pop(-1)

        with dest.open("w") as f:
            f.writelines(lines)


class YamlParser(BaseParser):
    @classmethod
    def read(cls, filename: Path) -> _ProtoNamespace:
        with filename.open("r") as f:
            res = yaml.load(f)
        return cls.from_conf(res)

    @classmethod
    def from_conf(cls, conf: tp.Mapping[str, tp.Any]) -> _ProtoNamespace:
        res = _ProtoNamespace()
        for key, val in conf.items():
            if key != "DEFAULT":
                res[key] = val

        cls.rec_walk(
            res,
            [
                (
                    lambda x: isinstance(x, CommentedMap),
                    lambda x: _ProtoNamespace(x.items()),
                ),
                (
                    lambda x: isinstance(x, CommentedSeq),
                    list,
                ),
            ],
        )
        return res

    @classmethod
    def to_conf(cls, config: _ProtoNamespace) -> tp.Any:
        return cls.rec_walk(
            config,
            [
                (
                    lambda x: isinstance(x, _ProtoNamespace),
                    lambda x: CommentedMap(x.items()) if x else None,
                ),
                (
                    lambda x: isinstance(x, list),
                    lambda x: CommentedSeq(x) if x else None,
                ),
            ],
        )

    @classmethod
    def write(cls, data: _ProtoNamespace, dest: Path) -> bool:
        cls.ensure_parent(dest)

        with tempfile.NamedTemporaryFile("w") as f:
            config = cls.to_conf(data)
            yaml.dump(config, f)

            if dest.exists() and filecmp.cmp(f.name, dest):
                return False

            shutil.copyfile(f.name, dest)
        return True

    @classmethod
    def write_stream(cls, data: _ProtoNamespace, f: tp.TextIO) -> None:
        config = cls.to_conf(data)
        yaml.dump(config, f)


class Jinja2Parser(BaseParser):
    @classmethod
    def read(cls, filename: Path) -> jinja2.Template:
        loader = [jinja2.FileSystemLoader(filename.parent)]

        # goes up 2 parents and checks for jinja-snippets dir
        loc = filename.parent
        for _ in range(2):
            for x in loc.iterdir():
                if not x.is_dir():
                    continue
                if x.name == "jinja-snippets":
                    loader.append(jinja2.FileSystemLoader(x))
            loc = loc.parent

        env = jinja2.Environment(
            loader=jinja2.ChoiceLoader(loader), undefined=jinja2.StrictUndefined
        )
        return env.get_template(str(filename.name))

    @classmethod
    def write(cls, data: str, dest: Path) -> bool:
        if data[-1:] != "\n":  # ensure newline
            data += "\n"

        while data[-2:] == "\n\n":
            data = data[:-1]

        cls.ensure_parent(dest)
        if dest.exists():
            with dest.open("r") as f:
                comp = f.read()
            if comp == data:
                return False

        with dest.open("w") as f:
            f.write(data)

        return True

    @classmethod
    @contextlib.contextmanager
    def render_to_tmp(cls, src: Path, **kwgs: tp.Any) -> tp.Iterator[Path]:
        tmp = tempfile.NamedTemporaryFile()
        path = Path(tmp.name)
        cls.render_to_dest(src, path, **kwgs)
        yield path

    @classmethod
    def render_to_dest(cls, src: Path, dest: Path, **kwgs: tp.Any) -> bool:
        assert ".jinja2" in src.suffix
        tpl = cls.read(src)
        render = tpl.render(**kwgs)
        return cls.write(render, dest)



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
# pylint: disable=too-many-lines
import typing as tp
from pathlib import Path

from .._parser import ConfigParser
from .._proto_namespace import _ProtoNamespace
from ._06_dependency import DependencyMixin


class MypyMixin(DependencyMixin):
    @classmethod
    def __keys(cls) -> tp.Tuple[str, ...]:
        return ("ignore",)

    def templated(self) -> None:
        super().templated()
        self.auxcon.dependencies.dev.append(self.versions.mypy)

    def demodata(self) -> None:
        super().demodata()
        self.auxcon.mypy = _ProtoNamespace(ignore=["click_help_colors"])

    def formatted(self) -> None:
        super().formatted()
        self._copy_keys_over(self.__keys(), "mypy")
        for key in self.__keys():
            self._to_list("mypy", key)

    def defaulted(self) -> None:
        super().defaulted()
        self.auxd.setdefault("mypy", _ProtoNamespace())
        for key in self.__keys():
            self.auxd.mypy.setdefault(key, [])

    def bake(self) -> None:
        super().bake()
        src = self.root / "pre-commit/mypy.ini"
        dest = self.target / "pre-commit/mypy.ini"
        config = ConfigParser.read(src)

        # namespace compat
        if "." in self.auxe.project.name:
            config["mypy"]["namespace_packages"] = True
            config["mypy"]["explicit_package_bases"] = True
            config["mypy"][
                "mypy_path"
            ] = f"$MYPY_CONFIG_FILE_DIR/../../{self.auxe.project.source_dir}"

        for x in self.auxe.mypy.ignore:
            config[f"mypy-{x}.*"] = _ProtoNamespace(ignore_missing_imports="True")

        # special django stubs case
        for dep in self.auxe.dependencies["dev"]:
            if (
                "django-stubs" in dep
                and (Path(self.auxe.project.source_dir) / "settings.py").exists()
            ):
                config["mypy"]["plugins"] = ["mypy_django_plugin.main"]
                config["mypy.plugins.django-stubs"] = dict(
                    django_settings_module=f"{self.auxe.project.name}.settings"
                )

        written = ConfigParser.write(config, dest)
        if written:
            self._print(f"baked {dest}", fg="green")



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
# pylint: disable=too-many-lines
import typing as tp

from .._parser import Jinja2Parser
from .._proto_namespace import _ProtoNamespace
from ._05_project import ProjectMixin


class GitIgnoreMixin(ProjectMixin):
    @classmethod
    def __keys(cls) -> tp.Tuple[str, ...]:
        return ("root", "source")

    def formatted(self) -> None:
        super().formatted()
        self._copy_keys_over(self.__keys(), "git_ignore")
        for key in self.__keys():
            self._to_list("git_ignore", key)

    def defaulted(self) -> None:
        super().defaulted()
        self.auxd.setdefault("git_ignore", _ProtoNamespace())
        for key in self.__keys():
            self.auxd.git_ignore.setdefault(key, [])

    def bake(self) -> None:
        super().bake()
        self.bake_file("gitignore", ".gitignore")

        data = self.auxe.project

        name1 = "root/gitignore"
        name2 = "root/source/gitignore"
        src1 = self.root / f"{name1}.jinja2"
        src2 = self.root / f"{name2}.jinja2"
        dest = self.target / ".." / ".gitignore"
        if data.source_dir != ".":
            self.bake_file(name1, "../.gitignore")
            self.bake_file(name2, f"../{data.source_dir}/.gitignore")

            flip = False
            for key in list(data.config):
                if flip:
                    data.config.move_to_end(key)
                if key == "options":
                    flip = True
                    data.config["options.packages.find"] = dict(where=data.source_dir)

        else:
            tpl1 = Jinja2Parser.read(src1)
            tpl2 = Jinja2Parser.read(src2)
            tpl = f"{tpl1.render(aux=self.auxe)}\n{tpl2.render(aux=self.auxe)}"
            written = Jinja2Parser.write(tpl, dest)
            if written:
                self._print(f"baked {dest}", fg="green")



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import dataclasses as dc
import typing as tp

from ._docker import DockerPayload

__all__ = ["DockerBuildPayload"]


from ._docker_executors import DockerBuild


@dc.dataclass
class DockerBuildPayload(DockerPayload):
    flavor: tp.ClassVar[str] = "docker_build"

    def create_executor(self, parents: tp.Any) -> tp.Any:
        return DockerBuild(
            slug=self.auxh.project.slug,
            service=self.param.service_name,
            image_name=self.param.image_name,
            branch_match=[],
            base_match=[],
            parents=parents,
            files=[f"devops/docker/{self.param.service_name}.dockerfile"],
        )

    def is_up_to_date(self) -> bool:
        return self.executor.is_up_to_date()  # type: ignore



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
# pylint: disable=too-many-lines
import typing as tp

from .._parser import Jinja2Parser
from .._parser import YamlParser
from .._proto_namespace import _ProtoNamespace
from ._03_meta import MetaMixin
from ._05_project import ProjectMixin
from ._06_dependency import DependencyMixin


class PrecommitMixin(DependencyMixin, ProjectMixin, MetaMixin):
    @classmethod
    def __keys(cls) -> tp.Tuple[str, ...]:
        return ("hooks", "rev_overwrite")

    def templated(self) -> None:
        super().templated()

        data = self.auxcon.dependencies
        data.dev.append(self.versions.pre_commit)
        data.dev_apt.append("git-core")

        data = self.auxcon.pre_commit = _ProtoNamespace(hooks=_ProtoNamespace())
        data.hooks["check-yaml"] = _ProtoNamespace(exclude="devops/CI")
        data.hooks["check-toml"] = _ProtoNamespace()
        data.hooks["check-json"] = _ProtoNamespace()
        data.hooks["end-of-file-fixer"] = _ProtoNamespace()
        data.hooks["add-copy-right"] = _ProtoNamespace()
        data.hooks["trailing-whitespace"] = _ProtoNamespace()
        data.hooks.black = _ProtoNamespace()
        data.hooks["blacken-docs"] = _ProtoNamespace()
        data.hooks.pyupgrade = _ProtoNamespace()
        data.hooks.pycln = _ProtoNamespace()
        data.hooks["reorder-python-imports"] = _ProtoNamespace()

        if self.is_enabled("mypy"):
            data.hooks.mypy = _ProtoNamespace()

        if self.is_enabled("pylint"):
            data.hooks.pylint = _ProtoNamespace()

    def demodata(self) -> None:
        super().demodata()
        data = self.auxcon.pre_commit
        data.hooks["pylint-test"] = dict(files="tests/")
        data.rev_overwrite = _ProtoNamespace(black="21.9b0")

    def formatted(self) -> None:
        super().formatted()
        self._copy_keys_over(self.__keys(), "pre_commit")
        for key in self.__keys():
            iter_mapping = key == "hooks"
            self._to_proto_ns("pre_commit", key, iter_mapping=iter_mapping)

    def defaulted(self) -> None:
        super().defaulted()
        self.auxd.setdefault("pre_commit", _ProtoNamespace())
        data = self.auxd.pre_commit
        for key in self.__keys():
            data.setdefault(key, _ProtoNamespace())

    def enriched(self) -> None:
        super().enriched()
        data = self.auxe.pre_commit
        custom_config = self.target_custom / "pre-commit" / "config.yaml"
        self.auxe.versions.update(
            {key: f"{key}=={val}" for key, val in data.rev_overwrite.items()}
        )
        data.custom = ""
        if custom_config.exists():
            with open(custom_config, encoding="utf-8") as f:
                data.custom = f.read()

    def bake(self) -> None:  # pylint: disable=too-many-branches
        super().bake()

        data = self.auxe.pre_commit
        srcj = self.root / "pre-commit/config.yaml.jinja2"
        with Jinja2Parser.render_to_tmp(srcj, aux=self.auxe) as src:
            config = YamlParser.read(src)

        requested_hooks = list(data.hooks.keys())

        self._raise_if_unsupported_hooks(config, requested_hooks)

        multi_hook_repo = ""
        # remove the ones not selected
        for repo in reversed(config.repos):

            def keep_selected(hook: _ProtoNamespace) -> bool:
                return hook.id in requested_hooks

            if len(repo.hooks) > 1:
                multi_hook_repo = repo.repo
            repo.hooks = list(filter(keep_selected, repo.hooks))
            if not repo.hooks:  # remove repo if empty
                config.repos.remove(repo)

            # integrate options
            for hook in repo.hooks:
                for key, val in data.hooks[hook.id].items():
                    if isinstance(val, list):
                        val = "|".join(val)
                    if key in ["coverage"]:
                        continue
                    if key not in ["files", "exclude"]:
                        raise NotImplementedError(
                            f"support for option '{key}' of {hook.id} not implemented yet"
                        )
                    if key in hook:
                        hook[key] += "|" + val
                    else:
                        hook[key] = val

        # check if local python files are required from the hook
        for repo in config.repos:
            for hook in repo.hooks:
                entry = hook.get("entry")
                if entry and entry.startswith("devops/") and "custom" not in entry:
                    self.bake_file(entry.replace("devops/", ""), chmod=0o755)

        # order the config according to the requested_hooks
        if multi_hook_repo != "":
            self._print(
                f"pre-commit: cannot sort repos: '{multi_hook_repo}' has multiple hooks",
                fg="red",
            )
        config.repos = sorted(
            config.repos, key=lambda x: requested_hooks.index(x.hooks[0].id)
        )

        dest = self.target / "pre-commit/config.yaml"
        written = YamlParser.write(config, dest)
        if written:
            self._print(f"baked {dest}", fg="green")

    @classmethod
    def _raise_if_unsupported_hooks(
        cls, config: _ProtoNamespace, requested_hooks: tp.Iterable[str]
    ) -> None:
        available_hooks: tp.List[str] = sum(
            ([hook.id for hook in repo.hooks] for repo in config.repos), []
        )
        unknown_hooks = set(requested_hooks) - set(available_hooks)
        if unknown_hooks:
            raise RuntimeError(
                f"pre-commit hooks are not supported by aux: {unknown_hooks}.\n"
                "       Hint: You could add them in the custom directory."
            )



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import enum

__all__ = ["ExtraLevel"]


class ExtraLevel(enum.Enum):
    FORMATTED = 0
    DEFAULTED = 1
    ENRICHED = 2
    HYDRATED = 3
    TEMPLATED = 4
    DEMODATA = 5



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import dataclasses as dc
import types
import typing as tp

from ..._proto_namespace import _ProtoNamespace
from ._base import Payload

__all__ = ["PythonPayload"]


@dc.dataclass
class PythonPayload(Payload):
    flavor: tp.ClassVar[str] = "python"
    fct: tp.Callable[
        [
            tp.Dict[str, tp.Any],
        ],
        bool,
    ]
    param: _ProtoNamespace = dc.field(default_factory=_ProtoNamespace)

    def run(self, force: bool = False) -> None:
        if hasattr(self, "_deps"):
            self.fct(self.auxh, self._deps, **self.param)  # type: ignore
        else:
            self.fct(self.auxh, **self.param)

    @classmethod
    def import_function(
        cls, name: str, custom_ppyload: tp.Optional[types.ModuleType]
    ) -> tp.Callable[[tp.Dict[str, tp.Any],], bool]:
        # pylint: disable=import-outside-toplevel
        from ...src.payload.python import functions as fallback

        fname = name.replace("-", "_")
        try:
            return getattr(custom_ppyload, fname)  # type: ignore
        except AttributeError:
            return getattr(fallback, fname)  # type: ignore

    def hydrate(self, deps: tp.Tuple["Payload", ...] = tuple()) -> None:
        if hasattr(self, "is_hydrated"):
            return
        # pylint: disable=attribute-defined-outside-init
        self.is_hydrated = True
        if deps:
            self._deps = deps



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
# pylint: disable=too-many-lines
import datetime
import os
import sys
import typing as tp
from pathlib import Path

from .._logging import logger
from .._parser import ConfigParser
from .._parser import Jinja2Parser
from .._proto_namespace import _ProtoNamespace
from ._02_base import BaseComponent


class ProjectMixin(BaseComponent):
    def templated(self) -> None:
        super().templated()
        data = self.auxcon.setdefault("project", _ProtoNamespace())
        data.creation_year = self.get_current_year()

    def demodata(self) -> None:
        super().demodata()
        data = self.auxcon.project
        data.name = "hallo.world"
        data.slug = "hlw"
        data.minimal_version = (self.deduce_python_version(),)
        data.author = "anonymous"
        data.license = "MIT"

    def formatted(self) -> None:
        super().formatted()
        keys = [
            "name",
            "slug",
            "author",
            "minimal_version",
            "supported_versions",
            "creation_year",
            "source_dir",
            "license",
            "project_urls",
        ]
        self._copy_keys_over(keys, "project")
        self._to_list("project", "project_urls")

    def defaulted(self) -> None:
        # pylint: disable=no-member
        super().defaulted()
        data = self.auxd.project
        defaults = dict(
            source_dir="source",
            license="Proprietary",
            project_urls=_ProtoNamespace(),
            supported_versions=[self.auxf.project.minimal_version],
        )
        for key, val in defaults.items():
            data.setdefault(key, val)

    def enriched(self) -> None:
        super().enriched()
        data = self.auxe.project
        data.minimal_version_slug = data.minimal_version.replace(".", "")
        data.module_dir = data.source_dir + "/" + data.name.replace(".", "/")
        data.namespace_name, data.second_name = "", data.name
        if "." in data.name:
            data.namespace_name, data.second_name = data.name.split(".", 1)
        data.active_years = self.get_active_years()
        data.setup_fields = [
            "name",
            "author",
            "license",
            "description",
            "long_description",
            "project_urls",
        ]
        data.version, _ = self.get_current_version_and_lines(data)
        data.release_notes = self.get_release_notes(data)

    def bake(self) -> None:
        super().bake()
        data = self.auxe.project
        self.bake_file("install-dev.sh", chmod=0o755)
        self.bake_file("root/_setup.py", "../setup.py")

        srcj = self.root / "root/setup.cfg.jinja2"
        with Jinja2Parser.render_to_tmp(srcj, aux=self.auxe) as src:
            data.config = ConfigParser.read(src)
            for key in data.setup_fields:
                if key not in data:
                    continue
                val = data[key]
                if val:
                    data.config.metadata[key] = val

        # license
        if data.license != "Proprietary":
            self.bake_file(
                f"license/{data.license}.txt",
                (self.target / ".." / "LICENSE.txt").resolve(),
            )

    def writeout(self) -> None:
        super().writeout()
        dest = self.target / "../setup.cfg"
        written = ConfigParser.write(self.auxe.project.pop("config"), dest)
        if written:
            self._print(f"baked {dest}", fg="green")

    @classmethod
    def deduce_project_name(cls, path: tp.Optional[Path] = None) -> str:
        path = path or (Path.cwd())

        # level 1
        for obj in path.glob("*/__init__.py"):
            if "__version__" in obj.open("r", encoding="utf-8").read():
                lvl1 = obj.parent.stem
                return lvl1
        # level 2
        for obj in path.glob("*/*/__init__.py"):
            if "__version__" in obj.open("r", encoding="utf-8").read():
                lvl1 = obj.parent.stem
                lvl2 = obj.parent.parent.stem
                if lvl2 in ["source"]:
                    return lvl1
                return f"{lvl2}.{lvl1}"
        return "not-found"

    @classmethod
    def deduce_project_slug(cls) -> str:
        proj_name = cls.deduce_project_name()
        if proj_name.count(".") == 1:
            ns, sub = proj_name.split(".")
            return ns[:2] + sub[:3]
        return proj_name[:3]

    @classmethod
    def deduce_python_version(cls) -> str:
        return ".".join(map(str, sys.version_info[:2]))

    @classmethod
    def deduce_user(cls) -> str:
        return os.environ.get("USER", os.environ.get("USERNAME", "unknown"))

    def get_active_years(self) -> str:
        current_year = self.get_current_year()
        creation_year = self.auxd.project.creation_year
        if creation_year == current_year:
            return f"{creation_year}"

        return f"{creation_year}-{current_year}"

    @classmethod
    def get_current_year(cls) -> str:
        return str(datetime.date.today().year)

    def get_current_version_and_lines(
        self, data: tp.Optional[_ProtoNamespace] = None
    ) -> tp.Tuple[str, tp.List[str]]:
        data = data or self.auxe.project
        dir_ = data.module_dir
        init = self.auxcon_file.parent / dir_ / "__init__.py"
        if not init.exists():
            logger.warning(
                "%s/%s does not exist, retuning 'n/a' as version!", dir_, init.name
            )
            return "n/a", []

        with init.open("r", encoding="utf8") as f:
            lines = f.readlines()
            for line in lines:
                if "__version__" in line:
                    version = line.strip().split('"', 2)[1]
                    break
            else:
                raise RuntimeError(f"version not found in {init}")

        return version, lines

    def get_release_notes(
        self, data: tp.Optional[_ProtoNamespace] = None
    ) -> tp.Dict[str, str]:
        data = data or self.auxd.project
        dir_ = data.source_dir
        notes = self.auxcon_file.parent / data.source_dir / "release-notes.txt"
        release_note: tp.Dict[str, str] = {}

        if not notes.exists():
            logger.warning(
                "%s/%s does not exist, retuning n/a as version!", dir_, notes.name
            )
            return {}
        with notes.open("r", encoding="utf8") as f:
            for line in f.readlines():
                version, note = line.strip().split(" ", 1)
                release_note[version] = note

        return release_note



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import copy
import typing as tp
from pathlib import Path

from .._parser import ConfigParser
from .._proto_namespace import _ProtoNamespace
from ._06_dependency import DependencyMixin


class PytestMixin(DependencyMixin):
    @classmethod
    def __keys(cls) -> tp.Tuple[str, ...]:
        return ("asyncio_mode", "markers", "addopts", "cache_dir")

    def templated(self) -> None:
        super().templated()
        data = self.auxcon.pytest = _ProtoNamespace()
        self.auxcon.dependencies.test.append(self.versions.pytest)
        data.markers = _ProtoNamespace({"merge_only": "run test only on merge request"})
        data.addopts = '--strict-markers -m "not merge_only"'

    def demodata(self) -> None:
        super().demodata()
        data = self.auxcon.pytest
        data.asyncio_mode = "strict"

    def formatted(self) -> None:
        super().formatted()
        self._copy_keys_over(self.__keys(), "pytest")
        self._to_proto_ns("pytest", "markers")

    def defaulted(self) -> None:
        super().defaulted()
        self.auxd.setdefault("pytest", _ProtoNamespace())
        self.auxd.pytest.setdefault("cache_dir", str(self.target / "cache" / "pytest"))
        self.auxd.pytest.setdefault("markers", _ProtoNamespace())

    def enriched(self) -> None:
        super().enriched()
        data = self.auxe.pytest
        if data.markers:
            data.markers = [f"{key}: {val}" for key, val in data.markers.items()]
        else:
            del data.markers

    def bake(self) -> None:
        super().bake()
        data = self.auxe.pytest
        project = self.auxe.project
        project.config["tool:pytest"] = data

        if project.source_dir == ".":
            # no need for pytest.ini, as setup.cfg handles it
            return

        # we need the ini, as pytest only mounts the source_dir
        # for tests, and then we dont have the setup information
        src_data = copy.deepcopy(data)
        for _ in project.source_dir.split("/"):
            src_data.cache_dir = Path("..") / src_data.cache_dir
        src_data.cache_dir = str(src_data.cache_dir)
        config = _ProtoNamespace(pytest=src_data)

        dest = self.target.parent / project.source_dir / "pytest.ini"
        written = ConfigParser.write(config, dest)
        if written:
            self._print(f"baked {dest}", fg="green")



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
# pylint: disable=too-many-lines
import contextlib
import typing as tp

from .._logging import logger
from .._proto_namespace import _ProtoNamespace
from ._00_extra_level import ExtraLevel
from ._02_base import BaseComponent


class SentinelMixin(BaseComponent):
    def __init__(self, *args: tp.Any, **kwgs: tp.Any) -> None:
        super().__init__(*args, **kwgs)
        self._auxcon_entered = False

    def bake(self) -> None:
        with self.extra():
            super().bake()
            self.writeout()

    def formatted(self) -> None:
        # idempotent guard
        if hasattr(self, "auxf"):
            return
        super().formatted()

    def defaulted(self) -> None:
        # idempotent guard
        if hasattr(self, "auxd"):
            return
        super().defaulted()

    def enriched(self) -> None:
        # idempotent guard
        if hasattr(self, "auxe"):
            return
        super().enriched()

    def hydrated(self) -> None:
        # idempotent guard
        if hasattr(self, "auxh"):
            return
        super().hydrated()

    @contextlib.contextmanager
    def extra(
        self, level: ExtraLevel = ExtraLevel.ENRICHED
    ) -> tp.Iterator[_ProtoNamespace]:
        logger.info("enabled %s", level)
        if level == ExtraLevel.TEMPLATED:
            self.templated()
            self.formatted()
            yield self.auxf
            del self.auxf
        elif level == ExtraLevel.DEMODATA:
            self.templated()
            self.demodata()
            self.formatted()
            yield self.auxf
            del self.auxf
        if level == ExtraLevel.FORMATTED:
            self.formatted()
            yield self.auxf
            del self.auxf
        elif level == ExtraLevel.DEFAULTED:
            self.formatted()
            self.defaulted()
            yield self.auxd
            del self.auxd
            del self.auxf
        elif level == ExtraLevel.ENRICHED:
            self.formatted()
            self.defaulted()
            self.enriched()
            yield self.auxe
            del self.auxe
            del self.auxd
            del self.auxf
        elif level == ExtraLevel.HYDRATED:
            self.formatted()
            self.defaulted()
            self.enriched()
            self.hydrated()
            yield self.auxh
            del self.auxh
            del self.auxe
            del self.auxd
            del self.auxf
        logger.info("disabled %s", level)



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
# pylint: disable=too-many-lines
import re
import typing as tp

from .._proto_namespace import _ProtoNamespace
from ._03_meta import MetaMixin
from ._05_project import ProjectMixin


class DependencyMixin(ProjectMixin, MetaMixin):
    def templated(self) -> None:
        super().templated()
        data = self.auxcon.dependencies = _ProtoNamespace()
        data.default = []
        data.test = []
        data.dev = []
        data.dev_apt = []

    def demodata(self) -> None:
        super().demodata()
        data = self.auxcon.dependencies
        data.default_apt = ["postgres"]
        data.default = ["numpy"]
        data.dev.pop(-1)

    def update_to_template(self, tpl: _ProtoNamespace, full: _ProtoNamespace) -> None:
        super().update_to_template(tpl, full)
        data = self.auxf.dependencies

        for mod in full.dependencies:
            if mod in tpl.dependencies and mod not in data:
                data[mod] = tpl.dependencies[mod]
                self._print(f"dependencies.{mod}: added {data[mod]}", fg="green")
            elif mod in data and mod not in tpl.dependencies:
                self._print(f"dependencies.{mod}: removed {data[mod]}", fg="red")
                del data[mod]

        for mod in tpl.dependencies:
            if mod not in ["test", "docs", "dev"]:
                continue
            newer = {}
            for dep in tpl.dependencies[mod]:
                pkg, version = self.parse_dep(dep)
                if version is not None:
                    newer[pkg] = version

            dep_list = data.get(mod, [])
            for i, dep in enumerate(dep_list):
                pkg, version = self.parse_dep(dep)
                if version is not None:
                    if pkg in newer and newer[pkg] != version:
                        self._print(
                            f"dependencies.{mod}: updated {pkg} {version}->{newer[pkg]}",
                            fg="green",
                        )
                        dep_list[i] = dep.replace(version, newer[pkg])

    @classmethod
    def __keys(cls) -> tp.Tuple[str, ...]:
        return ("default", "default_apt", "test", "test_apt", "dev", "dev_apt")

    def formatted(self) -> None:
        super().formatted()
        self._copy_keys_over(self.__keys(), "dependencies", allow_unknown=True)
        for key in self.__keys():
            self._to_list("dependencies", key)

    def defaulted(self) -> None:
        super().defaulted()
        self.auxd.setdefault("dependencies", _ProtoNamespace())
        data = self.auxd.dependencies
        for key in self.__keys():
            data.setdefault(key, [])

    @classmethod
    def parse_dep(cls, dep: str) -> tp.Tuple[str, tp.Optional[str]]:
        if "=" in dep:
            pkg, version = re.split("[=><~]{2}", dep, 1)
        else:
            pkg = dep
            version = None
        return pkg, version

    def bake(self) -> None:
        super().bake()
        config = self.auxe.project.config

        for key, deps in self.auxe.dependencies.items():
            if key.endswith("_apt") or key.endswith("_script"):
                continue
            if key == "default":
                # my config writer will mess up dependencies with environment markers
                # if ' are present, but I need ' in auxilium.cfg for the docker files
                # as some are "${VAR}" and '${VAR}' does not work
                config.options.install_requires = [x.replace("'", '"') for x in deps]
            elif deps:
                config["options.extras_require"][key] = deps



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
# pylint: disable=too-many-lines
import os
import typing as tp

from .._logging import logger
from .._proto_namespace import _ProtoNamespace
from .._util import ApiRequestCommunicator
from ._03_meta import MetaMixin
from ._18_payload import PayloadMixin


class CiMixin(PayloadMixin, MetaMixin):
    # pylint: disable=unused-private-member
    @classmethod
    def __keys(cls) -> tp.Tuple[str, ...]:
        return ("mechanism", "docker_image", "runner", "trigger", "use_adaux_img")

    def templated(self) -> None:
        super().templated()
        self.auxcon.ci = _ProtoNamespace(trigger=_ProtoNamespace())
        data = self.auxcon.ci.trigger
        data["+push"] = {
            "-openmr": {
                "-release": {"pre-commit": None, "pytest": None},
                "+release": {
                    "+gitlab": {
                        "gitlab-release": None,
                        "pkg-gitlab": None,
                        "pkg-pypi": None,
                    }
                },
            }
        }
        data["+mr"] = {
            "+draft": {
                "pre-commit": None,
                "pytest": None,
            },
            "-draft": {
                "pre-commit-all": None,
                "-vip": {
                    "pytest-mr": None,
                },
                "+vip": {
                    "pytest-cov-mr": None,
                },
            },
            "+release": {
                "check-release-notes": None,
            },
        }

    def demodata(self) -> None:
        super().demodata()
        self.auxcon.ci.runner = "dind-cached"
        self.auxcon.ci.mechanism = "mixed"

    def formatted(self) -> None:
        super().formatted()
        self._copy_keys_over(self.__keys(), "ci")

        for key in ["trigger"]:
            self._to_proto_ns("ci", key, iter_mapping=True)

    def defaulted(self) -> None:
        super().defaulted()
        self.auxd.setdefault("ci", _ProtoNamespace())
        data = self.auxd.ci

        data.setdefault("mechanism", "monolith")
        data.setdefault("runner", "normal")
        data.setdefault("docker_image", self.versions.ci_docker_image)
        data.setdefault("trigger", _ProtoNamespace())
        data.setdefault("use_adaux_img", True)
        assert data.trigger is not None
        assert data.mechanism in ["monolith"]
        assert data.runner in ["dind-cached", "normal"]

    # def update_to_template(self, tpl: _ProtoNamespace, full: _ProtoNamespace) -> None:
    #     super().update_to_template(tpl, full)

    #     data = self.auxd.ci.stages
    #     for key, stage in data.items():
    #         old = list(stage)
    #         new = list(tpl.ci.stages.get(key, {}))

    #         last_idx = 0
    #         for name, job in full.ci.stages.get(key, {}).items():
    #             if name in new and name not in old:
    #                 data.jobs.insert(last_idx + 1, job)
    #                 stage[name] = job
    #                 old.insert(last_idx + 1, name)
    #                 self._print(f"ci.jobs: added {name}", fg="green")
    #             elif name in old and name not in new:
    #                 del stage[name]
    #                 old.remove(name)
    #                 self._print(f"ci.jobs: removed {name}", fg="red")
    #             if name in old and name in new:
    #                 last_idx = old.index(name)

    def enriched(self) -> None:
        super().enriched()
        data = self.auxe.ci

        rule_used = _ProtoNamespace(
            mr=False,
            web=False,
            pipeline=False,
            schedule=False,
        )
        for reason in self._get_trigger_combos():
            for key in ["push", "push_no_mr"] + list(rule_used):
                if key in ["push", "push_no_mr"]:
                    if "+push" in reason:
                        rule_used.setdefault("push", False)
                        rule_used.setdefault("push_no_mr", True)
                        if "-openmr" not in reason:
                            rule_used["push_no_mr"] &= False
                            rule_used["push"] |= True
                else:
                    rule_used[key] |= f"+{key}" in reason

        data.used_rules = [key for key, val in rule_used.items() if val]

    # infuse shorthands
    # shorthand = dict(push=["push-no-mr", "mr"])
    # default_rules: tp.Dict[str, tp.List[str]] = {
    #     "rules": shorthand["push"] + ["web", "pipeline"],
    # }
    # for key in ["rules", "build-rules", "run-rules"]:
    #     res = []
    #     for rule in copy.deepcopy(val[key]):
    #         if rule in shorthand:
    #             val[key].remove(rule)
    #             res += shorthand[rule]
    #         else:
    #             res += [rule]
    #     if not val["rules"]:
    #         res = res or default_rules[key]
    #     val[key] = res

    # diff = set(val.keys()) - set(valid_opts_keys)
    # if diff:
    # raise RuntimeError(f"invalid options: {diff}")

    # data.setdefault("mechanism", "monolith")
    # data.setdefault("runner", "normal")
    # data.setdefault("docker_image", self.versions.ci_docker_image)
    # assert data.mechanism in ["monolith", "gitlab", "mixed"]
    # assert data.runner in ["dind-cached", "normal"]
    # yield auxcone

    def run_ci(self, trigger_str: str, dry: bool = False) -> int:
        if trigger_str == "gitlab":
            triggers = self._triggers_from_gitlab_ci()
        else:
            triggers = self._triggers_from_str(trigger_str)

        # get the payload_names
        payload_names = self._get_payload_names(triggers)

        self._print(f"triggers: {', '.join(triggers)}", fg="cyan")

        if not payload_names:
            self._print("no payloads selected", fg="yellow")
            return 0
        payloads = [self.auxh.payload.lookup[x] for x in payload_names]
        success = self.payload_run(*payloads, force=False, dry=dry)
        if success:
            if "+draft" in triggers:
                return 42
            return 0
        else:
            return 1

    def _triggers_from_gitlab_ci(self) -> tp.Sequence[str]:
        triggers = []

        def env(key: str) -> str:
            return os.environ.get(key, "")

        gitlab2adaux = {
            "push": "+push",
            "merge_request_event": "+mr",
            "web": "+web",
            "pipeline": "+pipeline",
            "schedule": "+schedule",
        }
        source_trigger = gitlab2adaux[env("CI_PIPELINE_SOURCE")]
        triggers.append(source_trigger)

        if env("CI_COMMIT_TAG") != "":
            assert source_trigger == "+push"
            source_trigger = "+tag"
        if env("CI_OPEN_MERGE_REQUESTS") != "":
            triggers.append("+openmr")
            # we dont care abound env("CI_OPEN_MERGE_REQUESTS") in push
            # draft is only an option for mr lines

        mr_iid = env("CI_MERGE_REQUEST_IID")
        if mr_iid != "":
            try:
                resp = self.get_mr_status(mr_iid)
                if resp["draft"]:
                    triggers.append("+draft")
                    logger.info("merge request %s is a draft", mr_iid)
                else:
                    logger.info("merge request %s is NOT a draft", mr_iid)
            except RuntimeError as err:
                self._print(
                    f"could not access gitlab api for checking mr draft ({err.args[0]})",
                    fg="red",
                )

        if source_trigger == "+push":
            branch = env("CI_COMMIT_BRANCH")
        elif source_trigger == "+mr":
            branch = env("CI_MERGE_REQUEST_TARGET_BRANCH_NAME")

        gitlab = self.auxh.gitlab
        if branch in gitlab.vip_branches:
            triggers.append("+vip")
        if branch == gitlab.default_branch:
            triggers.append("+default")
        if branch == gitlab.release_branch:
            triggers.append("+release")

        branch_trigger = f"+{branch}"
        if branch_trigger not in triggers:
            triggers.append(branch_trigger)

        triggers.append("+gitlab")
        return triggers

    def get_mr_status(self, mr_iid: str) -> tp.Dict[str, tp.Any]:
        coord = [
            "projects",
            os.environ["CI_PROJECT_ID"],
            "merge_requests",
            mr_iid,
        ]
        api = ApiRequestCommunicator()
        token = os.environ["GITLAB_READ_API"]
        api.headers = {"PRIVATE-TOKEN": token}
        api.base_url = "https://" + os.environ["CI_SERVER_HOST"]
        return api.api_request(*coord)  # type: ignore

    def _triggers_from_str(self, trigger_str: str) -> tp.Sequence[str]:
        if trigger_str[0] not in "+-":
            trigger_str = f"+{trigger_str}"

        triggers = []
        old = 0
        for i, char in enumerate(trigger_str):
            if char in "+-":
                triggers.append(trigger_str[old:i])
                old = i
        triggers.append(trigger_str[old:])
        waste = triggers.pop(0)
        assert waste == ""
        return triggers

    def _get_payload_names(self, triggers: tp.Sequence[str]) -> tp.Sequence[str]:
        res = []
        for payload_name, reason in self._get_payload_names_and_reason(triggers):
            logger.info("%s included due to %s", payload_name, "".join(reason))
            res.append(payload_name)
        return res

    def _get_payload_names_and_reason(
        self, triggers: tp.Sequence[str], collect_all: bool = False
    ) -> tp.Sequence[tp.Tuple[str, tp.Tuple[str, ...]]]:
        payload_names_reason = []
        data = self.auxe.ci.trigger

        def dig(data: _ProtoNamespace, reason: tp.Tuple[str, ...]) -> None:
            for key, val in data.items():
                if key.startswith("-"):
                    pos_key = key.replace("-", "+")
                    if pos_key not in triggers or collect_all:
                        dig(val, reason + (key,))
                elif key.startswith("+"):
                    if key in triggers or collect_all:
                        dig(val, reason + (key,))
                else:
                    payload_names_reason.append((key, reason))

        dig(data, tuple())

        return payload_names_reason

    def _get_trigger_combos(self) -> tp.Sequence[tp.Tuple[str, ...]]:
        reasons = set()
        for _, reason in self._get_payload_names_and_reason([], collect_all=True):
            reasons.add(reason)
        return list(reasons)

    def bake(self) -> None:  # pylint: disable=too-many-branches,too-many-locals
        super().bake()
        data = self.auxe.ci
        base_files = ["00-main.yml", "01-rules.yml"]
        assert data.mechanism == "monolith"

        for filename in base_files:
            self.bake_file(f"CI/{filename}")



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
# pylint: disable=too-many-lines
import typing as tp

from .._proto_namespace import _ProtoNamespace
from ._05_project import ProjectMixin
from ._06_dependency import DependencyMixin


class ExecutablesMixin(DependencyMixin, ProjectMixin):
    @classmethod
    def __keys(cls) -> tp.Tuple[str, ...]:
        return ("console_scripts", "scripts")

    def formatted(self) -> None:
        super().formatted()
        self._copy_keys_over(self.__keys(), "executables")
        for key in self.__keys():
            self._to_list("executables", key)

    def defaulted(self) -> None:
        super().defaulted()
        self.auxd.setdefault("executables", _ProtoNamespace())
        for key in self.__keys():
            self.auxd.executables.setdefault(key, [])

    def demodata(self) -> None:
        super().demodata()
        data = self.auxcon.executables
        data.scripts = ["scripts/say_hello"]

    def bake(self) -> None:
        super().bake()
        config = self.auxe.project.config
        data = self.auxe.executables

        config.options.scripts = data.scripts
        cscr = data.console_scripts
        for i, val in enumerate(cscr):
            if "=" in val:
                continue
            name = val.rsplit(":", 1)[1]
            cscr[i] = f"{name} = {val}"

        config["options.entry_points"].console_scripts = cscr



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
# pylint: disable=too-many-lines
import typing as tp

from .._proto_namespace import _ProtoNamespace
from ._05_project import ProjectMixin


class DocsMixin(ProjectMixin):
    @classmethod
    def __keys(cls) -> tp.Tuple[str, ...]:
        return ("strict", "framework", "root")

    def templated(self) -> None:
        super().templated()
        self.auxcon.dependencies.docs = [
            self.versions.sphinx,
            self.versions.sphinx_rtd_theme,
            self.versions.sphinx_click,
            self.versions.jupyter_sphinx,
            self.versions.bash_kernel,
        ]

    def demodata(self) -> None:
        super().demodata()
        data = self.auxcon.docs = _ProtoNamespace()
        data.root = "source/docs"

    def formatted(self) -> None:
        super().formatted()
        self._copy_keys_over(self.__keys(), "docs")

    def defaulted(self) -> None:
        super().defaulted()
        self.auxd.setdefault("docs", _ProtoNamespace())
        data = self.auxd.docs
        data.setdefault("root", f"{self.auxd.project.source_dir}/docs")
        data.setdefault("framework", "sphinx")

    def enriched(self) -> None:
        super().enriched()
        data = self.auxe.docs
        gitlab = self.auxd.gitlab
        if "url" not in data:
            pages_url = gitlab.remote_url.replace("gitlab", "pages")
            remote_name = gitlab.get("remote_name", self.auxe.project.second_name)
            data.setdefault(
                "url",
                f"https://{gitlab.remote_user}.{pages_url}/{remote_name}",
            )
        self.auxe.project.project_urls.Documentation = self.auxe.docs.url

    def bake(self) -> None:
        super().bake()
        data = self.auxe.docs

        # user docs dir
        dest_rel = f"../{data.root}"
        if data.framework == "sphinx":
            self.bake_file(
                "docs/user/conf.py", f"{dest_rel}/conf.py", only_if_inexistent=True
            )
            self.bake_file(
                "docs/user/index.rst", f"{dest_rel}/index.rst", only_if_inexistent=True
            )
            self.bake_file("docs/user/gitignore", f"{dest_rel}/.gitignore")

            for name in ["static"]:
                path = self.target / f"{dest_rel}/{name}"
                path.mkdir(parents=True, exist_ok=True)

            # devops docs dir
            dest_rel = "docs"
            for name in ["static", "templates"]:
                path = self.target / f"{dest_rel}/{name}"
                path.mkdir(parents=True, exist_ok=True)

            self.bake_file("docs/default_conf.py")
            self.bake_file("docs/postprocess_html.py")
            self.bake_file("docs/static/git-link-color.css")



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import collections
import typing as tp

from .._proto_namespace import _ProtoNamespace
from ._02_base import BaseComponent


class MetaMixin(BaseComponent):
    def formatted(self) -> None:
        super().formatted()
        keys = ["disabled"]
        self._copy_keys_over(keys, "meta")
        self._to_list("meta", "disabled")

    def defaulted(self) -> None:
        super().defaulted()
        self.auxd.setdefault("meta", _ProtoNamespace())
        self.auxd.meta.setdefault("disabled", [])

    def demodata(self) -> None:
        super().demodata()
        self.auxcon.meta = _ProtoNamespace(disabled=["docs"])

    def is_enabled(self, component_name: str) -> bool:
        for x in self.__class__.__mro__:
            if self._comp_name_type_match(component_name, x):
                return component_name not in self._disabled_list()
        return False

    def _disabled_list(self) -> tp.List[str]:
        return self.auxcon.get("meta", _ProtoNamespace()).get("disabled", [])  # type: ignore

    def type_wo_disabled(
        self,
        disabled_list: tp.Optional[tp.List[str]] = None,
        *,
        discard_before: str = "",
        check_absence: bool = True,
    ) -> "tp.Type[MetaMixin]":
        if disabled_list is None:
            disabled_list = self._disabled_list()
        res: tp.List[type] = []
        for part in self.__class__.__mro__:
            if part.__name__ == discard_before:
                res.clear()
            if part.__name__ in ["AllComponents", "DynComponent"]:
                continue
            if not any(self._comp_name_type_match(x, part) for x in disabled_list):
                res.append(part)
        res_type = BaseComponent.compose(*reversed(res))

        # check if disabled did not get added by enabled
        if check_absence:
            compare_to = self.type_wo_disabled(
                discard_before="SentinelMixin", check_absence=False
            )
            parents: tp.MutableMapping[
                type, tp.Sequence[type]
            ] = collections.defaultdict(list)
            for part in compare_to.__mro__[1:]:  # remove bottom dyn
                parents[part] = part.__mro__[1:]  # remove self == part
                if any(self._comp_name_type_match(x, part) for x in disabled_list):
                    used_by = [
                        key.__name__ for key, val in parents.items() if part in val
                    ]
                    raise RuntimeError(
                        f"{part.__name__} cannot be disabled, as it is used by {', '.join(used_by)}"
                    )

        return res_type  # type: ignore

    @classmethod
    def _comp_name_type_match(
        cls, component_name: str, component_type: tp.Type[tp.Any]
    ) -> bool:
        return cls._canon_comp_name(component_name) == cls._canon_type_name(
            component_type
        )

    @classmethod
    def _canon_type_name(cls, component_type: tp.Type[tp.Any]) -> str:
        return component_type.__name__.lower().replace("mixin", "")

    @classmethod
    def _canon_comp_name(cls, component_name: str) -> str:
        return component_name.replace("-", "")



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import copy
import typing as tp

from .._proto_namespace import _ProtoNamespace
from .._util import LazyVersionStr
from ._01_file_io_support import FileIOSupport


class BaseComponent(FileIOSupport):
    def __init__(self) -> None:
        super().__init__()
        self.versions = _ProtoNamespace(
            pytest="pytest>=7.2",
            pytest_cov="pytest-cov~=4.0",
            pre_commit="pre-commit>=2.20",
            mypy="mypy==1.1.1",
            pylint="pylint==2.17.1",
            black="black==23.1.0",
            sphinx="sphinx>=5.3.0",
            sphinx_rtd_theme="sphinx-rtd-theme>=1.0.0",
            sphinx_click="sphinx-click>=4.3",
            jupyter_sphinx="jupyter-sphinx>=0.4",
            bash_kernel="bash_kernel>=0.8",
            blacken_docs="blacken-docs==v1.12.1",
            pre_commit_hooks="pre-commit-hooks==v4.4.0",
            pyupgrade="pyupgrade==v3.3.1",
            pycln="pycln==v2.1.3",
            reorder_python_imports="reorder_python_imports==v3.9.0",
            isort="isort==5.12.0",
            encryption_check="encryption_check==v1.0.0",
            docstr_coverage="docstr-coverage==v2.2.0",
            requests="requests==2.28.2",
            types_requests="types-requests",
            adaux=f"adaux=={LazyVersionStr()}",
            docker_compose_file="3.8",
            ci_docker_image="docker:20.10.23",
            ci_adaux_image=f"prostructura/adaux:{LazyVersionStr()}",
        )
        self.auxcon = _ProtoNamespace()

    def formatted(self) -> None:
        # pylint: disable=attribute-defined-outside-init
        self.auxf = _ProtoNamespace()

    def defaulted(self) -> None:
        # pylint: disable=attribute-defined-outside-init
        self.auxd = copy.deepcopy(self.auxf)

    def templated(self) -> None:
        # uses auxcon directly
        pass

    def demodata(self) -> None:
        # uses auxcon directly
        pass

    def update_to_template(self, tpl: _ProtoNamespace, full: _ProtoNamespace) -> None:
        pass

    def enriched(self) -> None:
        # pylint: disable=attribute-defined-outside-init
        self.auxe = copy.deepcopy(self.auxd)
        self.auxe.versions = self.versions

    def hydrated(self) -> None:
        # pylint: disable=attribute-defined-outside-init
        self.auxh = copy.deepcopy(self.auxe)

    def bake(self) -> None:
        pass

    def writeout(self) -> None:
        pass

    @classmethod
    def compose(cls, *types: type) -> "tp.Type[BaseComponent]":
        if cls not in types:
            types = (cls,) + types
        bases = tuple(reversed(types))
        return type("DynComponent", bases, {})

    def _copy_keys_over(
        self,
        keys: tp.Sequence[str],
        cat: str,
        *subcats: str,
        allow_unknown: bool = False,
    ) -> None:
        scat, parent, data = self._drill_down((cat,) + subcats, self.auxcon)
        if scat is None:
            return

        res = _ProtoNamespace()
        for key in keys:
            if key in data:
                res[key] = copy.deepcopy(data[key])
        unknown = set(data) - set(keys)
        if unknown and allow_unknown:
            for key in filter(lambda x: x in unknown, data):
                res[key] = copy.deepcopy(data[key])
        unknown = set(data) - set(res)
        if unknown:
            raise RuntimeError(
                f"unknows options {unknown} for {'.'.join((cat,)+subcats)}"
            )

        if res:
            dest = self.auxf
            for scat in (cat,) + subcats:
                dest.setdefault(scat, _ProtoNamespace())
                parent = dest
                dest = dest[scat]

            if scat in parent:
                parent[scat].update(res)
            else:
                parent[scat] = res

    def _to_list(self, *cats: str) -> None:
        scat, parent, data = self._drill_down(cats, self.auxf)
        if scat is None:
            return
        parent[scat] = data or []

    def _to_proto_ns(
        self,
        *cats: str,
        iter_mapping: bool = False,
        source: tp.Optional[_ProtoNamespace] = None,
    ) -> None:
        source = source or self.auxf
        scat, parent, data = self._drill_down(cats, source)
        if scat is None:
            return
        parent[scat] = data = _ProtoNamespace(data or {})

        if iter_mapping:
            for key in data:
                self._to_proto_ns(key, source=data)

    def _drill_down(
        self, cats: tp.Tuple[str, ...], data: _ProtoNamespace
    ) -> tp.Tuple[tp.Optional[str], _ProtoNamespace, _ProtoNamespace]:
        cat = None
        for cat in cats:
            parent = data
            if data is None or cat not in data:
                return None, parent, parent
            data = data[cat]
        return cat, parent, data



# pylint: disable=relative-beyond-top-level
import os
import re
import typing as tp

from ...._components._payload._docker_executors import subprocess_run
from ...._components._payload._docker_executors import tag_image
from ...._components._payload._docker_executors import upload_to_remote
from ...._proto_namespace import _ProtoNamespace


def check_release_notes(aux: _ProtoNamespace) -> None:
    release_notes = aux.project.release_notes
    version = aux.project.version
    # check that version has a note
    if version not in release_notes:
        raise RuntimeError(f"version {version} is not in release notes, please add!")
    # check that version was not already released
    out = subprocess_run(
        ["git", "ls-remote", "--tags"], check=True, capture_output=True
    )
    already_released = list(re.findall(r"tags/([\d.]+)\n", out.stdout.decode()))
    if version in already_released:
        raise RuntimeError(
            f"version {version} was already released and cannot be released again!"
        )

    print(f"version {version} has description '{release_notes[version]}'")


def gitlab_release(aux: _ProtoNamespace) -> None:
    release_notes = aux.project.release_notes
    version = aux.project.version
    description = release_notes[version]
    os.environ["RELEASE_TAG"] = version
    os.environ["RELEASE_DESCRIPTION"] = description
    payload = aux.payload.lookup["gitlab-release-run"]
    payload.run()


def tag(
    aux: _ProtoNamespace,  # pylint: disable=unused-argument
    deps: tp.Any,
    tag: str,  # pylint: disable=redefined-outer-name
) -> None:
    if len(deps) != 1:
        raise RuntimeError(
            f"tag job for {tag} should have exactly 1 dependency, not {len(deps)}!"
        )
    tag = tag.format(version=aux.project.version)
    local_tag, release_tag = deps[0].executor.tag_and_upload(tag)
    print("uploaded", local_tag)
    print("   -> to", release_tag)


def img_dockerhub(
    aux: _ProtoNamespace,
    deps: tp.Any,
    release_tag: str,
) -> None:
    if len(deps) != 1:
        raise RuntimeError(
            f"img-dockerhub job for {release_tag} should have exactly 1 dependency, not {len(deps)}!"
        )
    release_tag = release_tag.format(
        version=aux.project.version, ci_adaux_image=aux.versions.ci_adaux_image
    )
    local_tag = deps[0].executor.pull_if_not_existent()
    tag_image(local_tag, release_tag)
    subprocess_run(
        [
            "docker",
            "login",
            "-u",
            os.environ["DOCKERHUB_USERNAME"],
            "-p",
            os.environ["DOCKERHUB_PASSWORD"],
            "docker.io",
        ]
    )
    upload_to_remote(local_tag, release_tag)
    subprocess_run(["docker", "logout", "docker.io"])
    print("uploaded", local_tag)
    print("   -> to", release_tag)



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import contextlib
import os
import sys
import time
import typing as tp
from pathlib import Path

import click
from click_help_colors import HelpColorsGroup

from ._cli_mixin import CliMixin
from ._cli_mixin import LazyVersionStr
from ._components import AllComponents
from ._logging import logging

__all__ = ["adaux", "hello"]


def hello(bla: str, param: tp.Optional[click.Parameter] = None) -> str:
    r"""
    Hello:

    .. include:: example.rst
    """
    print(param)
    return bla


class _ClickPrintMixin:
    def _print(self, msg: str, **kwgs: tp.Dict[str, tp.Any]) -> None:
        if self.verbose:  # type: ignore
            click.secho(msg, **kwgs)  # type: ignore


@click.group(
    cls=HelpColorsGroup, help_headers_color="yellow", help_options_color="green"
)
@click.version_option(version=LazyVersionStr())  # type: ignore
@click.pass_context
@click.option("--target", "-t", type=Path, help="Location of the repository.")
@click.option(
    "--bake-before",
    "-b",
    is_flag=True,
    default=False,
    help="Bake before executing the command.",
)
@click.option(
    "--silent",
    "-s",
    is_flag=True,
    default=False,
    help="Suppress any output on success.",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    default=False,
    help="Enables logging level Debug.",
)
def adaux(
    ctx: click.Context,
    target: tp.Optional[Path],
    bake_before: bool,
    silent: bool,
    verbose: bool,
) -> None:
    if verbose:
        logging.basicConfig(level=logging.DEBUG)
        logging.getLogger("urllib3").setLevel(logging.INFO)
    ctx.color = os.environ.get("NO_COLOR", "") == ""
    cls = AllComponents.compose(CliMixin, _ClickPrintMixin)
    bootstrap = cls(target, True)  # type: ignore
    with convert_runtime_to_click_error():
        ctx.obj = bootstrap.type_wo_disabled()(target, silent)  # type: ignore

    if ctx.obj.auxcon_file.exists():
        raise_if_init_and_not_force(ctx)
        if bake_before:
            ctx.invoke(bake)
    else:
        raise_if_not_init(ctx)


@adaux.command()
@click.pass_context
def demo(ctx: click.Context) -> None:
    r"""
    Print a more complex auxilium file for reference.
    """
    ctx.obj.demo()


@adaux.command()
@click.pass_context
@click.option(
    "-a", "--full", is_flag=True, default=False, help="display all extra settings"
)
def show(ctx: click.Context, full: bool) -> None:
    r"""
    Read and print auxilium file in yml syntax.
    """
    with convert_runtime_to_click_error():
        ctx.obj.show(full)


@adaux.command()
@click.pass_context
@click.option("-f", "--force", is_flag=True, default=False)
@click.option(
    "-n", "--project-name", prompt=True, default=AllComponents.deduce_project_name()
)
@click.option(
    "-s", "--project-slug", prompt=True, default=AllComponents.deduce_project_slug()
)
@click.option(
    "-p", "--python-version", prompt=True, default=AllComponents.deduce_python_version()
)
@click.option("-a", "--author", prompt=True, default=AllComponents.deduce_user())
def init(
    ctx: click.Context,
    project_name: str,
    project_slug: str,
    python_version: str,
    author: str,
    force: bool,
) -> None:
    r"""
    Initialize a new auxilium.aux.
    """
    with convert_runtime_to_click_error():
        ctx.obj.init(project_name, project_slug, python_version, author, force=force)


@adaux.command()
@click.pass_context
@click.option("--bake_after", "-b", is_flag=True)
def sync(ctx: click.Context, bake_after: bool) -> None:
    r"""
    Synchronizes local auxilium file with potentially newer template.
    """
    with convert_runtime_to_click_error():
        ctx.obj.sync()

    if bake_after:
        ctx.obj.set_defaults()
        ctx.obj.bake()


@adaux.command()
@click.pass_context
def bake(ctx: click.Context) -> None:  # pylint: disable=too-many-statements
    """
    Renders various files based on an auxilium file. Will overwrite without asking.
    """
    with convert_runtime_to_click_error():
        ctx.obj.bake()


@adaux.command()
@click.pass_context
@click.argument("auxcon_file")
def pre_commit_bake(ctx: click.Context, auxcon_file: str) -> None:
    assert "auxilium." in auxcon_file
    ctx.obj.bake()


gitlab_token_option = click.option(
    "-t",
    "--token",
    prompt=os.environ.get("GITLAB_API_TOKEN", "") == "",
    help="Gitlab Access Token with API Scope.",
    default="env:GITLAB_API_TOKEN",
)


@adaux.command()
@click.pass_context
@gitlab_token_option
def gitlab(ctx: click.Context, token: str) -> None:
    r"""
    Sets various settings on the remote gitlab repositroy.
    """
    token = _get_token_from_env(token)
    with convert_runtime_to_click_error():
        ctx.obj.gitlab(token)


@adaux.command()
@click.pass_context
@gitlab_token_option
@click.option(
    "-w",
    "--watch",
    default=False,
    is_flag=True,
    help="Check the pipeline status all 30 seconds.",
)
@click.option(
    "-a",
    "--show-success",
    is_flag=True,
    default=False,
    help="Show successful jobs.",
)
def pipeline(ctx: click.Context, token: str, watch: bool, show_success: bool) -> None:
    r"""
    Shows the most recend pipeline on the cli.
    """
    token = _get_token_from_env(token)
    with convert_runtime_to_click_error():
        while True:
            done, _ = ctx.obj.pipeline(token, show_success)
            if done or not watch:
                break
            time.sleep(30)


@adaux.command()
@click.pass_context
@click.option(
    "-M",
    "--major",
    is_flag=True,
    default=False,
    help="major tick",
)
@click.option(
    "-m",
    "--minor",
    is_flag=True,
    default=False,
    help="minor tick",
)
@click.option(
    "-c",
    "--commit",
    is_flag=True,
    default=False,
    help="commit with release message",
)
@click.argument("release_message", nargs=-1)
def tick(
    ctx: click.Context, release_message: str, major: bool, minor: bool, commit: bool
) -> None:
    r"""
    Tick the version and add the release note.
    """
    with convert_runtime_to_click_error():
        ctx.obj.tick(" ".join(release_message), major, minor, commit)


@adaux.command()
@click.pass_context
@gitlab_token_option
@click.option(
    "-w",
    "--watch",
    default=False,
    is_flag=True,
    help="Check the release status all 30 seconds.",
)
@click.option(
    "-p",
    "--pipeline-check",
    default=False,
    is_flag=True,
    help="Wait for successful pipeline status before executing release.",
)
@click.option(
    "-r",
    "--rerelease",
    default=False,
    is_flag=True,
    help="Allow the MR to have the same name as past closed MR.",
)
def release(
    ctx: click.Context, token: str, watch: bool, pipeline_check: bool, rerelease: bool
) -> None:
    r"""
    Create and display release merge request.
    """
    token = _get_token_from_env(token)
    with convert_runtime_to_click_error():
        success = not pipeline_check
        if pipeline_check:
            while True:
                done, success = ctx.obj.pipeline(token, show_success=False)
                if done or not watch:
                    break
                time.sleep(30)

        # will be skipped if pipeline is checked and failed
        while success:
            done, _ = ctx.obj.release(token, rerelease)
            if done or not watch:
                break
            time.sleep(30)


@adaux.command()
@click.pass_context
@click.option("-f", "--force", is_flag=True, default=False)
@click.argument("src", type=click.Path(exists=True), default=Path("auxilium.cfg"))
@click.argument("dest", type=Path, default=Path("auxilium.yml"))
def convert(ctx: click.Context, src: Path, dest: Path, force: bool) -> None:
    r"""
    Convert auxilium file into different format, e.g. cfg -> yml.
    """
    src = Path(src)
    with convert_runtime_to_click_error():
        ctx.obj.convert(src, dest, force)


def fwd_all_inputs(fct: tp.Any) -> tp.Any:
    return adaux.command(context_settings={"ignore_unknown_options": True})(
        click.pass_context(click.argument("args", nargs=-1)(fct))
    )


@adaux.command()
@click.pass_context
@click.argument("trigger_str", nargs=1)
@click.option("-d", "--dry", is_flag=True, default=False)
def ci(ctx: click.Context, trigger_str: tp.List[str], dry: bool) -> None:
    """
    Allows running the CI locally.
    """
    with convert_runtime_to_click_error():
        return_code = ctx.obj.ci(trigger_str, dry=dry)
    ctx.exit(return_code)


@adaux.command()
@click.pass_context
def cov(ctx: click.Context, open_html: bool) -> None:
    """
    Creates the coverage html report.
    """
    with convert_runtime_to_click_error():
        ctx.obj.cov(open_html)


@fwd_all_inputs
def dcp(ctx: click.Context, args: tp.List[str]) -> None:
    """
    Shortcut for docker compose with corresponding compose file.
    """
    with convert_runtime_to_click_error():
        ctx.obj.dcp(*args)


@adaux.command(context_settings={"ignore_unknown_options": True})
@click.pass_context
@click.argument("composition")
@click.argument("args", nargs=-1)
def comp(ctx: click.Context, composition: str, args: tp.List[str]) -> None:
    """
    Shortcut for specific compositions shell file: x-compose.sh
    """
    with convert_runtime_to_click_error():
        ctx.obj.comp(composition, *args)


@adaux.command()
@click.pass_context
@click.argument("payload_names", nargs=-1)
@click.option(
    "-f", "--force", is_flag=True, default=False, help="runs payload even if up to date"
)
def run(ctx: click.Context, payload_names: tp.Tuple[str, ...], force: bool) -> None:
    """
    Run a specified payload. Use 'ls' to list all available payloads.
    """
    with convert_runtime_to_click_error():
        ctx.obj.run(*payload_names, force=force)


@adaux.command()
@click.pass_context
@click.option("-h", "--open-html", is_flag=True, default=False)
def docs(ctx: click.Context, open_html: bool) -> None:
    """
    Creates the docs.
    """
    with convert_runtime_to_click_error():
        ctx.obj.docs(open_html)


@fwd_all_inputs
def mp(ctx: click.Context, args: tp.List[str]) -> None:
    """
    Shows the first lines of the mypy pre-commit hook.
    """
    with convert_runtime_to_click_error():
        ctx.obj.mp(*args)


@fwd_all_inputs
def pipi(ctx: click.Context, args: tp.List[str]) -> None:
    """
    Runs pip install with editable=strict option.
    """
    with convert_runtime_to_click_error():
        ctx.obj.pipi(*args)


@fwd_all_inputs
def pl(ctx: click.Context, args: tp.List[str]) -> None:
    """
    Shows the first lines of the pylint pre-commit hook.
    """
    with convert_runtime_to_click_error():
        ctx.obj.pl(*args)


@fwd_all_inputs
def pra(ctx: click.Context, args: tp.List[str]) -> None:
    """
    Runs all pre-commit hooks, or a specific one if supplied.
    """
    with convert_runtime_to_click_error():
        ctx.obj.pra(*args)


@adaux.command()
@click.pass_context
@click.option("-z", "--zipped", is_flag=True, default=False)
def sdist(ctx: click.Context, zipped: bool) -> None:
    """
    Builds the release package.
    """
    with convert_runtime_to_click_error():
        ctx.obj.sdist(zipped)


@adaux.command()
@click.pass_context
@click.option("-c", "--close", default=None)
@click.option("-i", "--gitignore", is_flag=True, default=False)
@click.argument("args", nargs=-1)
def todo(
    ctx: click.Context,
    args: tp.List[str],
    close: tp.Optional[str] = None,
    gitignore: bool = False,
) -> None:
    """
    Read and write todos.
    """
    with convert_runtime_to_click_error():
        new = " ".join(args) if args else None
        ctx.obj.todo_and_note("todo", new, close, gitignore)


@adaux.command()
@click.pass_context
@click.option("-c", "--close", default=None)
@click.option("-i", "--gitignore", is_flag=True, default=False)
@click.argument("args", nargs=-1)
def note(
    ctx: click.Context,
    args: tp.List[str],
    close: tp.Optional[str] = None,
    gitignore: bool = False,
) -> None:
    """
    Read and write notes.
    """
    with convert_runtime_to_click_error():
        new = " ".join(args) if args else None
        ctx.obj.todo_and_note("note", new, close, gitignore)


def _get_token_from_env(token: str) -> str:
    default = os.environ.get("GITLAB_API_TOKEN", "")
    if default == "":
        if token == "env:GITLAB_API_TOKEN":
            raise click.UsageError("GITLAB_API_TOKEN is not set!")
        click.secho(
            f"if you want to avoid re-entering, run: ' export GITLAB_API_TOKEN={token}'",
            fg="yellow",
        )

    if token == "env:GITLAB_API_TOKEN":
        token = default
    return token


@contextlib.contextmanager
def convert_runtime_to_click_error() -> tp.Iterator[None]:
    try:
        yield
    except RuntimeError as err:
        raise click.UsageError(err.args[0])


def raise_if_not_init(ctx: click.Context) -> None:
    if ctx.invoked_subcommand not in ["init"]:
        raise click.UsageError(
            f"{ctx.obj.auxcon_file} does not exists! use 'adaux init' to create one."
        )


def raise_if_init_and_not_force(ctx: click.Context) -> None:
    if ctx.invoked_subcommand in ["init"]:
        if all(x not in sys.argv for x in ["-f", "--force"]):
            raise click.UsageError(
                f"{ctx.obj.auxcon_file} already exists, Use 'adaux update' or -f to overwrite."
            )



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
from . import _00_extra_level
from . import _99_all
from ._00_extra_level import *
from ._99_all import *


__all__ = _99_all.__all__ + _00_extra_level.__all__
